/***************************************************************/ 
/*  Dekompressionsprogramm fr Selbstbau-Tauchcomputer (SBTC2) */ 
/*  Mikrocontroller: ATMEL AVR ATmega32, 8 MHz                 */ 
/*  Compiler:        GNU AVR C-Compiler                        */ 
/*  Autor:           Peter Rachow                              */ 
/*  Letzte nderung: 20:33 09.09.03                            */ 
/***************************************************************/ 

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

/************************/
/*   EEPROM-Speicher    */
/************************/
#include <avr/eeprom.h>
int bcount = 16;          /* Positionszeiger fr EEPROM */

/************************/
/*  Timer & Interrupt   */
/************************/
#define SCLOCK 8             /* Taktfrequenz im MHz */

volatile long loop1 = 0, loop1b = 0;
long int diveseconds = 0;

/************************/
/*      M I S C         */
/************************/
int main(void);
void led(char, char);
int round_depth(int);
int get_keys(void);
void settings(void);
void send_data(void);

void showtemp(void);
void show_gas(int);
void set_curgas(void);

void check_vasc(void);
void check_ppO2(void);
int maxppO2 = 16;  //1.6 bar

/************************/
/*      LCD-Display     */
/************************/
#define LCD_INST 0x00
#define LCD_DATA 0x01

void lcd_write(char, unsigned char, int);
void set_rs(char);
void set_e(char);
void lcd_init(void);
void lcd_cls(void);
void lcd_linecls(int);
void lcd_putchar(int, int, unsigned char);
void lcd_putstring(int, int, unsigned char*);
int lcd_putnumber(int, int, int, int, int);
void wait_ms(int);
void lcd_printdiveinfo(int, int, int);

/***************************/
/*       Dekorechnung      */
/***************************/
#define FN2 0.79        /* N2-Anteil im Atemgas    */


/* Gewebekonstanten fr 16 Kompartimente           */
/* STICKSTOFF                                      */
float t05N2[] = {4, 8, 12.5, 18.5, 27, 38.3, 54.3, 77, 109, 146, 187, 239, 305, 390, 498, 635};
float aN2[] = {1.2599, 1.0, 0.8618, 0.7562, 0.662, 0.5043, 0.441, 0.4, 0.375, 0.35, 0.3295, 
               0.3065, 0.2835, 0.261, 0.248, 0.2327};
float bN2[] = {0.505, 0.6514, 0.7222, 0.7825, 0.8126, 0.8434, 0.8693, 0.891, 
               0.9092, 0.9222, 0.9319, 0.9403, 0.9477, 0.9544, 0.9602, 0.9653};

/* Kompartimentsttigung */
float piN2[] = {0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72}; 

/* HELIUM */
float t05He[] = {1.51, 3.02, 4.72, 6.99, 10.21, 14.48, 20.53, 29.11, 41.2, 55.2, 70.7, 90.2, 115.2, 147.2, 188, 240};
float aHe[] = {1.7424, 1.383, 1.1919, 1.0458, 0.922, 0.8205, 0.7305, 0.6502, 
               0.595, 0.5545, 0.5333, 0.5189, 0.5181, 0.5176, 0.5172, 0.5119};
float bHe[] = {0.4245, 0.5747, 0.6527, 0.7223, 0.7582, 0.7957, 0.8279, 0.8553, 
               0.8757, 0.8903, 0.8997, 0.9073, 0.9122, 0.9171, 0.9217, 0.9267};

/* Kompartimentsttigung   */
float piHe[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; 
float pig[] = {0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72, 0.72}; 

/* 4 durch Anwender whlbare Gase (Gas1 = "Luft) */
unsigned char curgas = 0;
double figN2[4] = {FN2, 0, 0, 0};             /* N2-Anteil in 4 Auswahlgasen     */
double figHe[4] = {0, 0, 0, 0};               /* He-Anteil in in 4 Auswahlgasen  */

float airp = 0.995;             /* Umgebungsluftdruck in mbar */

int depth = 0, maxdepth = 0;    /* Akt. und max. Tiefe in dm  */
int ldepth = 0;                 /* Letzte aktuelle Teife fr v.asc */
int deepestdecostep = 0;        /* Tiefster Dekostopp in dm   */

int temp;                       /* Temperatur */

float get_water_pressure(int);
void calc_p_inert_gas(int);
float get_water_depth(float);
int calc_ndt(void);
void calc_deco(void);
void get_dsensor(void);
void get_tsensor(void);
void mod_abvalues(int);

/***************************/
/*        AD-Wandler       */
/***************************/
#define ADWAITSTATE 3
int adc_val;
char adc_mode = 0;

/******************************************************/
/*                    DEKO-Teil                       */
/******************************************************/
/******************************************************/
/*        Wasserdruck p.amb aus Tiefe depth           */
/*        berechnen                                   */
/******************************************************/
float get_water_pressure(int depth)
{
  return depth * 0.1 + airp;
}

/**********************************************************/
/*        Inertgaspartialdruck im Gewebe                  */
/*  d: Tiefe in m  Intervall immer 10 sec.                */    
/**********************************************************/
void calc_p_inert_gas(int d)
{
  unsigned char t1;
  float pamb = get_water_pressure(d) - 0.0627;

  for(t1 = 0; t1 <= 15; t1++)
  {
    piN2[t1] += (pamb * figN2[curgas]  - piN2[t1]) * (1 - exp((-1  * 0.1667 / t05N2[t1]) * log(2)));
    piHe[t1] += (pamb * figHe[curgas]  - piHe[t1]) * (1 - exp((-1  * 0.1667 / t05He[t1]) * log(2)));
    pig[t1] = piN2[t1] + piHe[t1];
  }
}

/*************************************************/
/*    Wassertiefe depth aus p.amb berechnen      */
/*************************************************/
float get_water_depth(float pamb)
{
  return (pamb - airp) * 10;
}

/**************************************************/
/*         Berechnen der Restnullzeit             */
/**************************************************/
int calc_ndt()
{
  char calcok = 0;      /* Flag, ob Rechnung OK ist */
  unsigned char t1; 

  int dp = depth * 0.1; /* Wassertiefe in m         */
  int t0min = 999;

  float te, xN2, xHe;
  float piigN2, piigHe, pamb = get_water_pressure(dp) - 0.0627;

  piigN2 = pamb * figN2[curgas];
  piigHe = pamb * figHe[curgas];
 
  for(t1 = 0; t1 < 16; t1++)
  {
    /*** Stickstoff ***/
    if(piigN2  - piN2[t1] && figN2[curgas])
    {
      xN2 = -1 * (airp / bN2[t1] + aN2[t1] - piN2[t1]) / (piigN2  - piN2[t1]) + 1;

      if(xN2 > 0) /* N2 Logarithmieren mglich? */
      {
        te = -1 * log(xN2) / log(2) * t05N2[t1]; 
        if(te < t0min)
	  t0min = te;
        calcok = 1;
      }
    }
    
    /*** Helium ***/
    if(piigHe - piHe[t1] && figHe[curgas])
    {
      xHe = -1 * (airp / bHe[t1] + aHe[t1] - piHe[t1]) / (piigHe - piHe[t1]) + 1;
     
      if(xHe > 0)
      {
        te = -1 * log(xHe) / log(2) * t05He[t1]; 
        if(te < t0min && te > 0)
	  t0min = te;
        calcok = 1;
      }
    }
  }
  
  if(!calcok || dp < 10)
    return -1;
  else
    if(t0min > 0)
      return t0min;
    else
      return 0;
}

/********************************************/
/*     Dekompressionsstufen berechnen       */
/********************************************/
void calc_deco()
{
  float piN2x[16], piHex[16], pigx[16];
  float pambtol, pambtolmax = 1.0;
  unsigned int decostep, dmin1 = 0, dmin2 = 0;
  unsigned char xpos = 0, t1;
  int ndt;

  /* Signal LED ein */
  led(2, 1);

  /* Aktuelle Gasspannungen in temporres Array bertragen */
  for(t1 = 0; t1 < 16; t1++)
  {
    piN2x[t1] = piN2[t1];
    piHex[t1] = piHe[t1];
    pigx[t1] = piN2x[t1] + piHex[t1];
  }

  for(t1 = 0; t1 < 16; t1++)        	              /* Erste Dekostufe */
    if((pigx[t1] - aN2[t1]) * bN2[t1] > pambtolmax)
      pambtolmax = (pigx[t1] - aN2[t1]) * bN2[t1];

  decostep = get_water_depth(pambtolmax);
  decostep = ((decostep / 3) + 1) * 3;

  deepestdecostep = 0;

  lcd_linecls(1);
  get_dsensor();

  while(decostep > 0)                       /* Nachfolgende Dekostufen bis 0 m Wassertiefe */
  {
    pambtolmax = 0.0;                       /* p.amb.tol.max berechnen */
  					    /* Inertgaspartialdruecke der Gewebe berechnen */
    for(t1 = 0; t1 < 16; t1++)
    {
      piN2x[t1] += ((get_water_pressure(decostep) - 0.0627) * figN2[curgas]  - piN2x[t1]) * (1 - exp((-1  * 1 / t05N2[t1]) * log(2)));
      piHex[t1] += ((get_water_pressure(decostep) - 0.0627) * figHe[curgas]  - piHex[t1]) * (1 - exp((-1 * 1 / t05He[t1]) * log(2)));
      pigx[t1] = piN2x[t1] + piHex[t1];
      if((pambtol = (pigx[t1] - aN2[t1]) * bN2[t1]) > pambtolmax)
	pambtolmax = pambtol;
    }

    /* vergleichen */
    if(get_water_depth(pambtolmax) < decostep - 3)
    {
      if(dmin1)
	xpos += lcd_putnumber(1, xpos, dmin1, -1, -1) + 1;

      decostep -= 3;
      dmin2 += dmin1;
      dmin1 = 0;
    }
    dmin1++;

    if(!deepestdecostep)
      deepestdecostep = decostep;
  }

  if(!dmin2)           /* Gesamte Dekozeit <= 0 also NZ-TG */
  {
    lcd_putstring(1, 0, "Nz.:");  

    ndt = calc_ndt();

    if(ndt < 0)       /* Unplausible NZ-Werte abfangen */
      lcd_putstring(1, 5, "---");      
    else
    {
      xpos = lcd_putnumber(1, 5, ndt, -1, -1) + 5;
      lcd_putchar(1, xpos, 39);
    }    
  }
  else                /* Dekompressionststopps erforderlich */
  {
    lcd_putchar(1, xpos++, 246);
    lcd_putchar(1, xpos++, '=');
    xpos += lcd_putnumber(1, xpos, dmin2, -1, -1);
    lcd_putchar(1, xpos, 39);
  }

  if(xpos < 12)
    showtemp();

  /* Signal LED aus */
  led(2, 0);
}

/******************************************************/
/*       bersttigungstoleranzen verndern           */
/******************************************************/
void mod_abvalues(int k)
{

  unsigned char old_k = eeprom_rb(10);
  unsigned char t1;
  double f = 1.3 - (old_k * 0.1);
 
  //Eventuell alte Werte wiederherstellen
  if(old_k > 0 && old_k < 8)
  {
    for(t1 = 0; t1 < 15; t1++)
    {
      aN2[t1] /= f;
      bN2[t1] *= f;
      aHe[t1] /= f;
      bHe[t1] *= f;
    }
  }

  f = 1.3 - (k * 0.1);
  for(t1 = 0; t1 < 15; t1++)
  {
    aN2[t1] *= f;
    bN2[t1] /= f;
    aHe[t1] *= f;
    bHe[t1] /= f;
  }
  eeprom_wb(10, k);
}

/******************************************************/
/*            Aufstiegsgeschwindigkeit                */
/******************************************************/
void check_vasc()
{
  int v;

  v = (ldepth - depth) * 6;

  if(v > 10)
    led(3, 1);    
  else
    led(3, 0);    
}  

/******************************************************/
/*                        ppO2                        */
/******************************************************/
void check_ppO2()
{
  float lppO2 = 10 * ((depth / 100) + 1) * (1 - figN2[curgas] - figHe[curgas]);
  
  if(lppO2 > maxppO2)
  {
    lcd_linecls(1);
    lcd_putstring(1, 0, "!! ppO2 !!");
  }
}

/******************************************************/
/*               Drucksensor auslesen                 */
/******************************************************/
void get_dsensor()
{
  unsigned char xpos;

  adc_mode = 0;

  /* AD-Wandler */
  outp(64 + 0, ADMUX); /* AD-Wandler: Interne Referenz und Kanal 0 aktivieren PA0 PIN 40 */
  wait_ms(ADWAITSTATE);
  outp(206, ADCSRA);   /* AD-Wandler abfragen */

  if(depth > maxdepth)
    maxdepth = depth;
     
  lcd_printdiveinfo(depth, maxdepth, diveseconds * 0.0166666667);

  if(depth < deepestdecostep * 10)
  {
    led(4, 1);
    lcd_putstring(0, 6, "!  ");
    xpos = lcd_putnumber(0, 7, deepestdecostep, -1, -1) + 7;                        
    lcd_putstring(0, xpos, "m! ");
    wait_ms(50);
    led(4, 0);
  }
}

/******************************************************/
/*             Temperatursensor auslesen              */
/******************************************************/

void get_tsensor()
{
  adc_mode = 1;

  /* AD-Wandler */
  outp(64 + 1, ADMUX); /* AD-Wandler: Interne Referenz und Kanal 1 aktivieren PA1 PIN 39 */
  wait_ms(ADWAITSTATE);
  outp(206, ADCSRA);   /* AD-Wandler abfragen */
}


/*** Ende Dekoteil ***********************************************************/



/******************************************************/
/*                  L C D -Teil                       */
/******************************************************/

/*********************************************************/
/*   Ein Byte (Befehl bzw. Zeichen) zum Display senden   */
/*********************************************************/
void lcd_write(char lcdmode, unsigned char value, int waitcycles)
{

  set_e(0); 

  if(!lcdmode)
    set_rs(0);    //RS=0 => Befehl
  else
    set_rs(1);    //RS=1 => Zeichen

  wait_ms(waitcycles * 2);

  set_e(1);
  outp(value & 0xF0, PORTD);           /* Hi byte */
  set_e(0);
  set_e(1);
  outp((value & 0x0F) * 0x10, PORTD);  /* Lo byte */
  set_e(0);

}

/******************************************************/
/* Ein Zeichen (Char) zum Display senden, dieses in   */
/* Zeile row und Spalte col positionieren             */
/******************************************************/
void lcd_putchar(int row, int col, unsigned char ch)
{
  lcd_write(LCD_INST, col + 128 + row * 0x40, 1);
  lcd_write(LCD_DATA, ch, 1);
}

/******************************************************/
/*    Eine Zeichenkette direkt in das LCD schreiben   */
/*    Parameter: Startposition, Zeile und Pointer     */
/******************************************************/
void lcd_putstring(int row, int col, unsigned char *s)
{
  unsigned char t1;

  for(t1 = col; *(s); t1++)
    lcd_putchar(row, t1, *(s++));
}


/******************************************************/
/*                  Display lschen                   */
/******************************************************/
void lcd_cls(void)
{
  lcd_write(LCD_INST, 1, 5);
}


/******************************************************/
/*       Display lschen (eine Zeile)                 */
/******************************************************/

void lcd_linecls(int displine)
{
  unsigned char t1;

  for(t1 = 0; t1 <= 15; t1++)
    lcd_putchar(displine, t1, 32);
}


/**********************************************************/
/*                    RS und E setzen                     */
/**********************************************************/
void set_e(char status)  /* PORT C0 = Pin 6 am LCD */
{
  if(status)
    sbi(PORTC, 0);
  else
    cbi(PORTC, 0);
}

void set_rs(char status) /* PORT C1 = Pin 4 am LCD */
{
  if(status)
    sbi(PORTC, 1);
  else
    cbi(PORTC, 1);
}

/*************************************************/
/*         LCD-Display initialisieren            */
/*************************************************/
void lcd_init(void)
{
  /* Grundeinstellungen: 2 Zeilen, 5x7 Matrix, 4 Bit */
  lcd_write(LCD_INST, 40, 5);  
  lcd_write(LCD_INST, 40, 5);
  lcd_write(LCD_INST, 40, 5);

  lcd_write(LCD_INST, 2, 5);
  lcd_write(LCD_INST, 8, 5);

  /* Display on, Cursor off, Blink off */
  lcd_write(LCD_INST, 12, 5);

  lcd_cls();

  /* Entrymode !cursoincrease + !displayshifted */
  lcd_write(LCD_INST, 4, 5);
}


/*************************************************************/
/*    Eine x-stellige Zahl direkt in das LCD schreiben       */
/*    Parameter: Startposition und Zeile; Zahl,              */
/*    darzustellende Ziffern, Position des Dezimalpunktes    */
/*************************************************************/
int lcd_putnumber(int row, int col, int num, int digits, int dec)
{
  char cl = col;
  int t1, t2, n = num, r, x = 1;
  
  /* Stellenzahl automatisch bestimmen */
  if(digits == -1)
  {
    for(t1 = 1; t1 < 10 && (n / x); t1++)
      x *= 10;
    digits = t1 - 1;
  }

  if(!digits)
    digits = 1;

  for(t1 = digits - 1; t1 >= 0; t1--)
  {
    x = 1;
    for(t2 = 0; t2 < t1; t2++)
      x *= 10;
    r = n / x;
    lcd_putchar(row, cl++, r + 48);
    if(t1 == dec) 
      lcd_putchar(row, cl++, 46); /* Dezimalpunkt */
    n -= r * x;
  }

  return digits;
}

/*******************************************************/                     
/*    Alle Daten an die richtigen Stellen des LCD      */                     
/*    schreiben: Parameter: Tiefen in m, Zeit in sec.  */                     
/*******************************************************/                     
void lcd_printdiveinfo(int cdepth, int mdepth, int divetime)
{ 
  char xpos = 0;

  lcd_putnumber(0, 0, cdepth, 3, 1);                        

  lcd_putchar(0, 4, 'm');

  if(mdepth > 999)
    mdepth = 990;

  lcd_putchar(0, 6, '(');
  lcd_putnumber(0, 7, round_depth(mdepth), 2, -1);                        
  lcd_putstring(0, 9, "m)");

  xpos = lcd_putnumber(0, 12, divetime, -1, -1) + 12;
  lcd_putchar(0, xpos, 39);
}                                                                             

/*************************************************/
/*            Temperatur anzeigen                */
/*************************************************/
void showtemp()
{
  unsigned char xpos;

  if(temp >= 0)
  {
    xpos = lcd_putnumber(1, 13, temp, -1, -1) + 13;
    lcd_putchar(1, xpos,  223);
  }
}

/*************************************************/
/*     Aktuelles Atemgas anzeigen und setzen     */
/*************************************************/
void show_gas(int gasnum)
{
  unsigned char xpos;

  lcd_cls();
  lcd_putstring(0, 0, "Gas"); 
  lcd_putnumber(0, 4, gasnum + 1, -1, -1);

  lcd_putstring(1, 0, "N2:");
  xpos = lcd_putnumber(1, 4, figN2[gasnum] * 100, -1, -1) + 4;
  lcd_putchar(1, xpos, '%');

  lcd_putstring(1, 8, "He:");
  xpos = lcd_putnumber(1, 12, figHe[gasnum] * 100, -1, -1) + 12;
  lcd_putchar(1, xpos, '%');
}    

/*************************************************/
/*        Aktuelles Atemgas einstellen           */
/*************************************************/
void set_curgas()
{
  unsigned char lcurgas = curgas;
  long int lseconds = diveseconds;

  show_gas(lcurgas);

  while(get_keys());

  do
  {
    if(get_keys() == 3)
    {
      lseconds = diveseconds;
      lcurgas++;
      if(lcurgas > 3)
        lcurgas = 0;
      show_gas(lcurgas);  
      while(get_keys() == 3);
    }    
  }
  while(diveseconds < lseconds  + 5);

  lcd_cls();

  if(curgas != lcurgas)
  {
    lcd_putstring(0, 0, "Wechsel zu Gas");
    lcd_putnumber(0, 15, lcurgas + 1, -1, -1);
    curgas = lcurgas;
  }
  else
    lcd_putstring(0, 0, "Kein Gaswechsel"); 

  wait_ms(2000);
  lcd_cls();
}
     

/******************************************************/
/* Tiefenwert runden auf 1 Nachkommastelle u. /= 10   */
/******************************************************/
int round_depth(int d)
{
  if(d - (d / 10) * 10 >= 5)
    return (d / 10 + 1);
  else
    return (d / 10);
}

/*** Ende LCD-Teil ***********************************************************/


/***************************************************/
/* Wartezeit in Millisekunden bei fck = 8.000 MHz  */
/***************************************************/
void wait_ms(int ms)
{
  int t1, t2;

  for(t1 = 0; t1 < ms; t1++)
    for(t2 = 0; t2 < 137 * SCLOCK; t2++)
      asm volatile ("nop" ::);
}

/*************************************************/
/*             LEDs an PC2, 3 und 4              */
/*************************************************/
void led(char lednum, char status)
{
  if(status)
    cbi(PORTC, lednum);
  else
    sbi(PORTC, lednum);
}      


/***************************************************/
/*             Timer 0 Ereignisroutine             */
/***************************************************/
SIGNAL(SIG_OVERFLOW0)
{
  outp(0, TCNT0);       /* Timerregister auf 0 */

  if(loop1++ >= 486)    /* Sekundenzhler inkrementieren originaler Wert = 486 = Echtzeit */
  {
    diveseconds++;
    loop1 = 0;
  }
  loop1b++;
}

/***************************************************/
/*           AD-Wandler Ereignisroutine            */
/***************************************************/
SIGNAL(SIG_ADC)
{
  unsigned char lo, hi;

  lo = inp(ADCL);
  hi = inp(ADCH);

  adc_val = hi * 256 + lo;
  if(!adc_mode)
    depth = adc_val ;
  else
    temp = ((adc_val) - 594) / -6,24;
}

/*************************************************/
/*        Einstellungen durch Benutzer           */
/*  Tastentprellung per Hardware erforderlich    */
/*************************************************/
int get_keys(void)
{
  if(inp(PINB) & 1) //PB 0 
    return 1; 

  if(inp(PINB) & 2) //PB 1  
    return 2;

  if(inp(PINB) & 4) //PB 2
    return 3;
  
  return 0;
}

/*************************************************/
/*           Benutzereinstellungen               */
/*************************************************/
void settings(void)
{
  int menu_airp = airp * 1000;
  int menu_altp = menu_airp;
  int menu_ppO2 = maxppO2; 
  int menu_N2[4];
  int menu_He[4];
  int menu_h = 0, intv, t1;
  char menu_cons = eeprom_rb(10);
  char menu_cons2 = eeprom_rb(10);
  char ch, xpos;

  for(t1 = 0; t1 < 4; t1++)
  {
    menu_N2[t1] = figN2[t1] * 100;
    menu_He[t1] = figHe[t1] * 100;
  }

  while(get_keys());
  lcd_cls();

  lcd_putstring(0, 0, "Luftdruck NN"); 
  xpos = lcd_putnumber(1, 0, menu_airp, -1, -1);
  lcd_putstring(1, xpos, "mbar");
  do
  {
    ch = get_keys();
    if(ch == 1 || ch == 3)
    {
      lcd_linecls(1);
      intv =  menu_airp / 5;
      switch(ch)
      {
        case 3: menu_airp = intv * 5 + 5;
                if(menu_airp > 1100)
	          menu_airp = 900;
                break;
        case 1: menu_airp = intv * 5 - 5;
                if(menu_airp < 900)
	          menu_airp = 1100;
      }
      xpos = lcd_putnumber(1, 0, menu_airp, -1, -1);
      lcd_putstring(1, xpos, "mbar");
      wait_ms(100);
    }
  }while(ch != 2);
  while(get_keys());
  lcd_cls();

  lcd_putstring(0, 0, "Hoehe ueber NN"); 
  xpos = lcd_putnumber(1, 0, menu_h, -1, -1);
  lcd_putstring(1, xpos, "m");
  do
  {
    ch = get_keys();
    if(ch == 1 || ch == 3)
    {
      lcd_linecls(1);

      switch(ch)
      {
        case 3: menu_h += 100;
                if(menu_h > 4000)
           	  menu_h = 0;
                break;
        case 1: menu_h -= 100;
                if(menu_h < 0)
           	  menu_h = 4000;
                break;
      }

      xpos = lcd_putnumber(1, 0, menu_h, -1, -1);
      lcd_putstring(1, xpos, "m");
      lcd_putchar(1, 6, '(');
      menu_altp = menu_airp * exp(-1.2928 * 0.0981 * menu_h / menu_airp);
      xpos = lcd_putnumber(1, 7, menu_altp, -1, -1) + 7;
      lcd_putstring(1, xpos, "mbar)");
      wait_ms(100);
    }
  }while(ch != 2);

  if(menu_h > 0)
    menu_airp = menu_altp;

  while(get_keys());
  lcd_cls();

  //4 Atemgase
  // N2
  for(t1 = 0; t1 < 4; t1++)
  {
    lcd_putstring(0, 0, "Gas   N2-Anteil"); 
    lcd_putnumber(0, 4, t1 + 1, -1, -1); 
    xpos = lcd_putnumber(1, 0, menu_N2[t1], -1, -1);
    lcd_putstring(1, xpos, "%  ");
    do
    {
      ch = get_keys();
      if(ch == 1 || ch == 3)
      {
        lcd_linecls(1);
        switch(ch)
        {
          case 3: menu_N2[t1]++;
                  if(menu_N2[t1] > 79)
	            menu_N2[t1] = 0;
                  break;
          case 1: menu_N2[t1]--;
                  if(menu_N2[t1] < 0)
	            menu_N2[t1] = 79;
        }
        xpos = lcd_putnumber(1, 0, menu_N2[t1], -1, -1);
        lcd_putstring(1, xpos, "%  ");
        lcd_putstring(1, 5, "Nitrox");
        lcd_putnumber(1, 12, 100 - menu_N2[t1], -1, -1);
      }
      wait_ms(100);
    }while(ch != 2);

    lcd_linecls(1);

    lcd_putstring(0, 0, "Gas   He-Anteil"); 
    lcd_putnumber(0, 4, t1 + 1, -1, -1); 
    xpos = lcd_putnumber(1, 0, menu_He[t1], -1, -1);
    lcd_putstring(1, xpos, "%  ");
    do
    {
      ch = get_keys();
      if(ch == 1 || ch == 3)
      {
        lcd_linecls(1);
        switch(ch)
        {
          case 3: menu_He[t1]++;
                  if(menu_He[t1] > 99)
	            menu_He[t1] = 0;
                  break;
          case 1: menu_He[t1]--;
                  if(menu_He[t1] < 0)
	            menu_He[t1] = 99;
        }
        xpos = lcd_putnumber(1, 0, menu_He[t1], -1, -1);
        lcd_putstring(1, xpos, "%  ");
      }
      wait_ms(100);
    }while(ch != 2);

    while(get_keys());
    lcd_cls();
  }


  //ppO2
  lcd_putstring(0, 0, "Maximaler ppO2");
  xpos = lcd_putnumber(1, 0, menu_ppO2, -1, 1);
  lcd_putstring(1, xpos + 1, "bar");  

  do
  {
    ch = get_keys();
    if(ch == 1 || ch == 3)
    {
      switch(ch)
      {
        case 3: menu_ppO2 += 1;
                if(menu_ppO2 > 20)
                  menu_ppO2 = 10;
                break;
        case 1: menu_cons -= 1;
                if(menu_ppO2 < 10)
                  menu_ppO2 = 20;
      }
      lcd_putnumber(1, 0, menu_ppO2, -1, 1);
      while(get_keys());
      wait_ms(100);
    }
  }while(ch != 2);
  while(get_keys());
  lcd_cls();

  //Konservativ-Faktor
  lcd_putstring(0, 0, "Konservativ-");
  lcd_putstring(1, 0, "faktor:"); 
  lcd_putnumber(1, 8, menu_cons, -1, -1);
    
  do
  {
    ch = get_keys();
    if(ch == 1 || ch == 3)
    {
      switch(ch)
      {
        case 3: menu_cons += 1;
                if(menu_cons > 6)
                  menu_cons = 1;
                break;
        case 1: menu_cons -= 1;
                if(menu_cons < 1)
                  menu_cons = 6;
      }
      xpos = lcd_putnumber(1, 8, menu_cons, -1, -1);
      while(get_keys());
      wait_ms(100);
    }
  }while(ch != 2);
  while(get_keys());
  lcd_cls();

  // Datenbertragung zum PC starten?
  lcd_putstring(0, 0, "Tr");
  lcd_putchar(0, 3, 126);
  lcd_putstring(0, 5, "PC? (j/n)");  
  do
  {
    if(get_keys() == 3)
    {
      cli();
      send_data();
      sei();      
    }
  }while(get_keys() != 2);
  while(get_keys());
  lcd_cls();

  // Speichern?
  lcd_putstring(0, 0, "Sichern? (j/n)"); 
  do
  {
    if(get_keys() == 3)
    {
      cli();
      //Luftdruck am Tauchort
      airp = menu_airp * 0.001;
      eeprom_wb(0, menu_airp & 0xFF);            //LoByte
      eeprom_wb(1, (menu_airp & 0xFF00) / 256);  //HiByte
      while(!eeprom_is_ready());

      //Gase
      for(t1 = 0; t1 < 4; t1++)
      {
        figN2[t1] = menu_N2[t1] * 0.01;
        figHe[t1] = menu_He[t1] * 0.01;
        eeprom_wb(t1 * 2 + 2, menu_N2[t1]);
        eeprom_wb(t1 * 2 + 3, menu_He[t1]);
      }
      while(!eeprom_is_ready());

      //Konservativfaktor
      if(menu_cons != menu_cons2 && menu_cons)
      {
        mod_abvalues(menu_cons);
        eeprom_wb(10, menu_cons);
      }

      //maxppO2
      maxppO2 = menu_ppO2;
      eeprom_wb(11, menu_ppO2);

      sei();
      lcd_putstring(1, 7, "OK.");
    }
  }while(get_keys() != 2);

  lcd_cls();
}

/*************************************************/
/*   Inhalt des EEPROM seriell an PB2 ausgeben   */
/*************************************************/
void send_data(void)
{

  int t1;
  unsigned char t2;
  unsigned char b, x;

  lcd_putstring(1, 5, "Bytes");

  for(t1 = 0; t1 < 1024; t1++)
  {
    x = 128;
    b = eeprom_rb(t1);
    lcd_putnumber(1, 0, t1 + 1, -1, -1);
    for(t2 = 0; t2 < 8; t2++)
    {
      sbi(PORTB, 2);
      if(b & x)
        wait_ms(4);
      else  
        wait_ms(2);
      cbi(PORTB, 2);
      wait_ms(2);
      x /= 2;
    }
    wait_ms(2);
  }
}

/*************************************************/
/*      Start- und Hauptprogramm main()          */
/*************************************************/
int main()
{
  long int seconds_old, seconds_old2;
  int t1;

  //Konservativ-Faktor auf 4 setzen *= 1.1
  mod_abvalues(4);

  /* Ports einrichten */  
  /* OUTPUT */
  /* Port D Bit 4-7 auf "Output" schalten (= DB4-DB7)       */
  outp(0xF0, DDRD);
  /* Port C Bit 0 bis 4 auf "Output" schalten (= RS, E und LEDs) */
  outp(0x1F, DDRC);

  /* INPUT */
  /* Port A komplett */
  outp(0x00, DDRA);  

  /* Mit LCD-Initialisierung 0.5 s warten bis PowerUp von MC OK */
  wait_ms(500);
  lcd_init();                  
  lcd_putstring(0, 1, "SBTC2 (ZH-L16)");
  lcd_putstring(1, 1, "V2.0.40");
  wait_ms(2000);
  lcd_cls();

  //Alle LEDs aus
  for(t1 = 2; t1 < 5; t1++)
    led(t1, 0);

  /* Grundeinstellungen aus EEPROM lesen */
  lcd_putstring(0, 0, "Einstellungen:"); 

  //Umgebungsluftdruck
  airp = (eeprom_rb(0) + eeprom_rb(1) * 256) * 0.001;  //Luftdruck
  if(airp < 0.66 || airp > 1.2)
    airp = 1;
  lcd_putstring(1, 0, "Luftdruck"); 
  lcd_putnumber(1, 10, airp * 1000, -1, -1);

  wait_ms(1000);

  // N2- und He-Anteile in den 4 Gasen 
  for(t1 = 0; t1 < 4; t1++)
  {
    figN2[t1] = eeprom_rb(t1 * 2 + 2) * 0.01;
    if(figN2[t1] < 0 || figN2[t1] > 0.79)
      figN2[t1] = 0.78;

    figHe[t1] = eeprom_rb(t1 * 2 + 3) * 0.01;
    if(figHe[t1] < 0 || figHe[t1] > 0.99)
      figHe[t1] = 0;

    show_gas(t1);
    wait_ms(1000);
  }
  curgas = 0;

  lcd_cls();

  //maxppO2
  maxppO2 = eeprom_rb(11);
  if(maxppO2 > 20 || maxppO2 < 10)
    maxppO2 = 16;
  
  lcd_putnumber(1, 0, maxppO2, -1, 1);  
  lcd_putstring(0, 1, "Max. ppO2");
  lcd_putstring(1, 3, "bar");

  wait_ms(1000);

  lcd_cls();

  /* Timer 0 fr Sekundenzhlung */
  outp((1<<TOIE0), TIMSK); /* Overflow Interrupt = 1           */
  outp(0, TCNT0);          /* Startwert T/C0                   */
  outp(3, TCCR0);          /* Echtzeit: Taktrate 3 = (ck / 64) */

  sei();
  
  seconds_old = diveseconds - 10;
  seconds_old2= diveseconds - 1;

  for(;;) /* Endlosschleife fr Druckmessung und Dekorechnung */
  {
    /* Sensorabfrage ca. alle 0.25 sec. */
    if(loop1b > 250)
    {
      get_dsensor();
      loop1b = 0;
    }

    if(diveseconds >= seconds_old2 + 1)    
    {
      check_ppO2();
      check_vasc();
      ldepth = depth;
      seconds_old2 = diveseconds;
    }

    /* Alle 10 sec. Gewebesttigung & Dekorechnung */
    if(diveseconds >= seconds_old + 10)    
    {  
      get_tsensor();
      calc_p_inert_gas(depth * 0.1);
      calc_deco();
      seconds_old = diveseconds;

      cli(); /* Interrupts deaktivieren solange EEPROM geschrieben wird */
      if(bcount < 1024)          
        eeprom_wb(bcount++, depth * 0.1); 
      sei();
    }

    /* Tastaturabfrage ob Einstellungen gesetzt werden sollen */
    switch(get_keys())
    {
      case 2: settings();      
              break;
      case 3: set_curgas();
    }
  }
  return 0;
}

